/* $Id: dyn_falib.c,v 1.8 2001/03/07 19:32:47 jm Exp $
 * Dynamics Foreign Agent API support library implementation
 *
 * Dynamic hierarchial IP tunnel
 * Copyright (C) 1998-2000, Dynamics group
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation. See README and COPYING for
 * more details.
 */

#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <unistd.h>
#include <string.h>

#include "lib_api.h"
#include "dyn_falib.h"



#define TMP_PREFIX   "falib"
#define TMP_DIR       NULL
#define SOCK_PERM     0600

#define MIN(x,y) (((x) < (y)) ? (x) : (y))

#define DEFAULT_PATH  "/var/run/dynamics_fa_admin"


/*
 * Initialize library
 * Parameters:
 *   agent - path to mobility agent
 * Returns:
 *  API_FAILED - error in intialization
 *  API_SUCCESS - success
 *   
 */

int dynamics_fa_init(char *agent)
{  
	if (agent != NULL) 
		return lib_init(agent, TMP_DIR, TMP_PREFIX, SOCK_PERM);
	else return lib_init(DEFAULT_PATH, TMP_DIR, TMP_PREFIX, SOCK_PERM);
}

int dynamics_fa_get_tunnels(int *tunnel_count, dyn_tunnel_id *data,
			    int timeout)
{
	int r;
	struct api_msg msg;

	if (tunnel_count == NULL ||
	    *tunnel_count <= 0 ||
	    data == NULL ||
	    timeout < -1)
		return API_ILLEGAL_PARAMETERS;

	r = perform_call(API_GET_TUNNELS, NULL, 0, timeout, &msg);
  
	if (r != API_SUCCESS) return r;

	r = MIN(msg.length / sizeof(dyn_tunnel_id), *tunnel_count);
	*tunnel_count = msg.length / sizeof(dyn_tunnel_id);
	memcpy(data, msg.params, r * sizeof(dyn_tunnel_id));

	return (r == *tunnel_count)? API_SUCCESS : API_INSUFFICIENT_SPACE;
  
}
int dynamics_fa_get_tunnel_info(dyn_tunnel_id id, 
				struct dynamics_tunnel_info *data, 
				int *size_of_data, int timeout)
{
	int r;
	struct api_msg msg;

	if (data == NULL ||
	    size_of_data == NULL ||
	    *size_of_data <= 0 ||
	    timeout < -1)
		return API_ILLEGAL_PARAMETERS;
  
	r = perform_call(API_GET_TUNNEL_INFO, &id, 
			 sizeof(dyn_tunnel_id), timeout, &msg);
  
	if (r != API_SUCCESS) return r;

	r = MIN(*size_of_data, msg.length);
  
	memcpy(data, msg.params, r);
	*size_of_data = msg.length;

	return API_SUCCESS;  
}
int dynamics_fa_destroy_tunnel(dyn_tunnel_id id, int timeout)
{
	int r;
	struct api_msg msg;

	if (timeout < -1) 
		return API_ILLEGAL_PARAMETERS;

	r = perform_call(API_DESTROY_TUNNEL, &id, 
			 sizeof(dyn_tunnel_id), timeout, &msg);
  
	return r;
}

int dynamics_fa_get_status(struct dynamics_fa_status *status, int timeout)
{
	int r;
	struct api_msg msg;

	/* check parameters */
	if (status == NULL || timeout < -1)
		return API_ILLEGAL_PARAMETERS;

	/* do the call */
	r = perform_call(API_GET_STATUS, NULL, 0, timeout, &msg);

	if (r != API_SUCCESS) return r;

	if (msg.length != sizeof(struct dynamics_fa_status))
		return API_FAILED;

	memcpy(status, msg.params, msg.length); 

	return API_SUCCESS;

}

char *dynamics_fa_get_error_string(int err)
{
	return get_error_string(err);
}

/*
 * Get current agent address
 */
char *dynamics_fa_get_agent_path(void)
{
	return get_agent_path();
}

int dynamics_fa_attach_socket_info(char *path)
{
	struct api_msg msg;

	if (path == NULL) return API_ILLEGAL_PARAMETERS;
	return perform_call(API_ATTACH_INFO_SOCKET, path,
			    strlen(path) + 1, -1, &msg);
}

int dynamics_fa_detach_socket_info(char *path)
{
	struct api_msg msg;

	if (path == NULL) return API_ILLEGAL_PARAMETERS;
	return perform_call(API_DETACH_INFO_SOCKET, path,
			    strlen(path) + 1, -1, &msg);
}

int dynamics_fa_send_adv(struct dyn_param_adv *adv)
{
	struct api_msg msg;
	if (adv == NULL)
		return API_ILLEGAL_PARAMETERS;
	return perform_call(API_SEND_ADV, adv, 
			    sizeof(struct dyn_param_adv), -1, &msg);
}

