/* $Id: dyn_wireless.h,v 1.10 2001/06/18 16:16:25 jm Exp $
 * Dynamics wireless module header file
 *
 * Dynamic hierarchial IP tunnel
 * Copyright (C) 1998-2000, Dynamics group
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation. See README and COPYING for
 * more details.
 */

#ifndef DYN_WIRELESS_H
#define DYN_WIRELESS_H

#include <linux/if_ether.h>

/* Directly from the driver. Keep consistent with it (see the patch)!! */
#define MAX_NUM_QUALS 64
/* Private ioctls */
#define SET_HISTORY SIOCDEVPRIVATE+0xd
#define GET_HISTORY SIOCDEVPRIVATE+0xe

#define SET_QUALITIES SIOCDEVPRIVATE+0x3
#define GET_QUALITIES SIOCDEVPRIVATE+0x4
#define CHECK_QUALITIES SIOCDEVPRIVATE+0x5

struct quality_values {
	int num_of_quals;             /* how many entries used in this node */
	int sim_quals[MAX_NUM_QUALS]; /* every node contains at maximum */
	unsigned long sim_timestamps[MAX_NUM_QUALS]; /* timestamp relative 
							to the start */
	struct quality_values *next;  /* next node, if more quality values */
};

int dyn_wireless_get_ifname(const char *str, char *ifname);
int dyn_wireless_create_socket(void);
int dyn_wireless_iwspy_get(int sock, char *ifname, char *buffer);
int dyn_wireless_iwspy_set(int sock, char *ifname, char *buffer, 
			   int monitored);
int dyn_wireless_his_get(int sock, char *ifname, char *buffer);
int dyn_wireless_his_set(int sock, char *ifname, char *range, int size);
struct quality_values *dyn_wireless_get_quals_node(struct quality_values *q);
struct quality_values *dyn_wireless_get_simulator(int sock, char *ifname, 
						  char *hw);
int dyn_wireless_set_simulator(int sock, char *ifname, char *hw, 
			       struct quality_values *q, int msec_ival,
			       int scale);
int dyn_wireless_check_simulator(int sock, char *ifname, char *hw); 
int dyn_wireless_set_channel(int sock, char *ifname, int channel);
int dyn_wireless_get_channel(int sock, char *ifname);
int dyn_wireless_get_name(int sock, char *ifname, char *name, int len);
int dyn_wireless_get_nwid(int sock, char *ifname);
int dyn_wireless_set_nwid(int sock, char *ifname, char *id, int len);
int dyn_wireless_get_range(int sock, char *ifname, struct iw_range *range);
int dyn_wireless_get_ap_address(int sock, const char *ifname, char *hw);

#endif
