/* $Id: msgparser.h,v 1.24 2001/07/13 17:07:29 jm Exp $
 * Parsing of Registration Request and Reply Messages
 *
 * Dynamic hierarchial IP tunnel
 * Copyright (C) 1998-2001, Dynamics group
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation. See README and COPYING for
 * more details.
 * 
 * Additional modifications Copyright 2010 Juho Paaso, Aalto 
 * University / Department of Communication and Networking"
 */

#ifndef MSGPARSER_H
#define MSGPARSER_H

#include "message.h"

#define DOUBLE_MH_AUTH 0x01
#define DOUBLE_MF_AUTH 0x02
#define DOUBLE_FH_AUTH 0x04
#define DOUBLE_FF_AUTH 0x08
#define DOUBLE_SK_AUTH 0x10
#define DOUBLE_SHA_HA_AUTH 0x20

struct msg_extensions {
	char *start;
	int len;
	
	struct reg_req *req;
	struct reg_rep *rep;
	
	/* UDP Tunnel Extensions */
	struct udp_tnl_data 		*tnl_data;
	struct udp_tnl_req_ext 		*tnl_req;
	struct udp_tnl_rep_ext 		*tnl_rep;
	
	/* Network Mobility (NEMO) Extensions */
	struct mobile_nw_req_ext	*nemo_req[NEMO_NW_MAX_AMOUNT]; /* Array of pointers to a struct */
	struct mobile_nw_ack_ext	*nemo_ack[NEMO_NW_MAX_AMOUNT];
	int nemo_nw_amount;

	/* Home Agent assisted Route Optimization (HAaRO) Extensions */
	struct ro_capab_ext		ro_capab;
	struct ro_reply_ext		*ro_reply;
	struct mr_cr_auth_ext	*mr_cr_auth;
	struct coa_ext			*coa_ext;
	struct ro_prefix_adv	ro_prefix_adv;
	
	struct msg_auth *mh_auth;
	struct msg_auth *mf_auth;
	struct msg_auth *fh_auth;
	struct vendor_msg_auth *ff_auth;
	struct vendor_msg_auth *sk_auth;
	struct msg_key *fa_keyreq;
	struct msg_key *fa_pubkey;
	struct msg_key *mn_keyreq;
	struct msg_key *mn_keyrep;
	struct msg_key *fa_keyrep;
	struct msg_key *fa_pubkeyrep;
	struct registration_ext_dynamics *ext_dyn;
	struct msg_key *pubkey_hash;
	struct encaps_delivery_ext *encaps_del;
	struct fa_nai_ext *fa_nai;
	struct fa_nai_ext *prev_fa_nai;
	struct gre_key_ext *gre_key;
	struct sfa_debug_ext *sfa_debug;
	struct fa_reg_req *fa_req;
	struct fa_reg_rep *fa_rep;
	struct mn_nai_ext *mn_nai;
	struct priv_ha_ext *priv_ha;
	struct vendor_msg_auth *sha_ha_auth;
	struct nonce_ext *nonce;
	struct challenge_ext *challenge;
	struct generalized_auth_ext *mn_aaa_auth;
	struct generalized_mn_fa_key_req_ext *mn_fa_key_req_aaa;
	struct generalized_mn_fa_key_rep_ext *mn_fa_key_material_aaa;
	struct generalized_mn_ha_key_req_ext *mn_ha_key_req_aaa;
	struct generalized_mn_ha_key_rep_ext *mn_ha_key_material_aaa;

	/* bitfield of duplicate authentication extensions */
	unsigned char double_auth_ext;
	int unknown_cvse; /* RFC 3025: unknown critical vendor specific ext. */

	unsigned short mr_port;	/* in nw byte order */
	int registration_socket;
};

int parse_msg(char *msg_start, int len, struct msg_extensions *ext);

#endif
